#!/usr/bin/env python

from ds485d_mock import *
from ds485c_mock import *
from notification_queue import *
from process_pkt import *
from ds485.dsuid import *
from util import list2hexstr
from iptables import *
from testsuite_harness import *

from Queue import Queue, Empty
import unittest
import time, datetime
import sys, os
import traceback, signal
import logging
import asyncore, socket
import random
from threading import Thread, Timer

TIMESTMP_FMT = ("%H:%M:%S") # no %f for msecs
FORMAT = '%(asctime)s.%(msecs)03d %(name)-5s %(message)s'
logging.basicConfig(format=FORMAT, datefmt=TIMESTMP_FMT, level=logging.INFO)

class DelayReply(Thread):
    def __init__(self, delay, src, reply):
        Thread.__init__(self)
        self.delay = delay
        self.src = src
        self.reply = reply

    def run(self):
        log = logging.getLogger("DelayReplay")
        log.info("send device list in %d [s]", self.delay)
        time.sleep(self.delay)
        log.info("send device list reply")
        self.src.push(bytearray(self.reply))


class Hrack(unittest.TestCase):
    log = logging.getLogger("testsuite")

    def setUp(self):
        log = self.log
        #log.setLevel(logging.DEBUG)
        log.debug("setUp")
        log.debug("PID %d", os.getpid())
        signal.signal(signal.SIGINT, self.signal_handler)

        try:
            ports = [ 5004, 5005 ]
            self.ds485_port = 8445

            os.system("pkill -9 ds485p")
            for p in ports:
                iptables_unblock(log, p)

            fork_ds485p(self.ds485_port, ports,
                        loglevel = 5, valgrind = True)

            self.select_loop = AsyncCore()
            self.select_loop.start()

            # open the socket after forking
            self.daemons = fire_up_daemons(ports)
            if not self.daemons:
                log.info("")
                log.info("ds485p failed to connected to our fake-buses")
                log.info("")
                seppuku()


            log.info("")
            log.info("ds485p connected to our fake-buses")
            log.info("")

        except Exception, e:
            print traceback.format_exc()
            raise e # TODO is tearDown called?


    def signal_handler(self, signal, frame):
        self.log.warning('You pressed Ctrl+C!')
        seppuku()

    def tearDown(self):
        self.log.info("tearDown")
        kill_ds485p()
        asyncore.close_all()
        for d in self.daemons:
            d.close()
        self.select_loop.stop()

    def stop(self):
        self.log.info("do stop")
        self.run = False

    #@unittest.skip("")
    def test_dsm_reboots(self):
        log = self.log
        log.info("")
        log.info("query device <10s")
        log.info("")
        log.info("dsm client has a timeout of 10s when querying devices list.")
        log.info("Usual clients immediately refetch the device list after bus change")
        log.info("Is received. Up to now ds485p is doing the same. Ther problem is")
        log.info("ds485p forwards the bus_change upon reception then queries it's busses")
        log.info("until it is done with that query it will not reply to clients, which")
        log.info("they will time out.")
        log.info("Setup: ds485p sandwiched between 2 busses and 1 client")
        log.info("For each bus change received query the device list of the ds485p, then")
        log.info("stop watch until reply arrives. If this exceeds 10s test failed.")
        log.info("Whole process is started by emitting random(0,5) bus_change via mocked")
        log.info("sub busses. Query device requests of the ds485p are delayed by a")
        log.info("random[5, 10]s timeout")
        log.info("")

        dsmids = []
        dsmids_to_src = {}
        for d in self.daemons:
            dsmids.extend(d.state.dsmids)
            for dsm in d.state.dsmids:
                dsmids_to_src[dsm] = d

        class states:
            initialize = 'i'
            send_new_bc = 's'
            wait_bc = 'w'
            check_fetch_devices = 'g'

        tstate = states.initialize
        off = []
        pending_msg = []
        ct = 0
        cutoff = 10 # max messages
        countdown = 0
        client = None
        need_fetch = 0
        device_list_queries = []
        device_list_replies = []

        self.run = True
        try:
            while self.run:
                try:
                    (evt, pkt, src) = [None] * 3
                    (evt, pkt, src) = notification_queue.get(True, 1)
                    reply = None
                except Empty: # timeout

                    countdown -= 1
                    if countdown > 0:
                        continue

                    if not client:
                        log.info("Start Client")
                        client = Client('localhost', self.ds485_port, ClientState())
                        continue

                    cstate = client.state

                    if tstate == states.initialize:
                        log.info("initialize")
                        if not cstate.connected:
                            client.push(bytearray(createQueryBusState(0)))
                            continue

                        # wait 20s for initial bus changes from sysConnThread
                        countdown = 20
                        cstate.dsmids = list(dsmids) # hack, dsmids_iteratively
                        tstate = states.send_new_bc
                        continue

                    elif tstate == states.send_new_bc:

                        log.info("send_new_bc")

                        pending_msg = []
                        device_list_queries = []
                        device_list_replies = []

                        for i in [1]: #range(random.randint(0, 5)):
                            ct += 1
                            cur = dsmids[random.randint(0, len(dsmids) - 1)]
                            if cur in off:
                                off.remove(cur)
                                m = createBusChange(cur.octets, 0)
                            else:
                                off.append(cur)
                                m = createBusChange(cur.octets, 1)

                            pending_msg.append((cur, cur not in off))
                            log.info("bus_change%d: %s dsuid: %s", ct,
                                     "leave" if (cur in off) else "join ", cur)

                            d = dsmids_to_src[cur]
                            if cur in off:
                                d.state.dsmids.remove(cur)
                            else:
                                d.state.dsmids.append(cur)
                            d.send_all_clients(bytearray(m))

                        tstate = states.wait_bc
                        continue

                    elif tstate == states.wait_bc:
                        log.info('wait query devices replies')

                        log.info(device_list_replies)
                        log.info(pending_msg)
                        if (len(device_list_replies) != len(pending_msg)):
                            continue

                        log.info("device list queries delays")
                        for (start, end) in zip(device_list_queries,
                                                device_list_replies):
                            log.debug("\tstart %r end %r", end, start)
                            log.info("\tdevice query: delay %r [s]", (end - start).seconds)
                            self.assertTrue((end -start).seconds <= 10)

                        log.info("query device list done")
                        tstate = states.check_fetch_devices

                        countdown = 7 # give it some rest

                    elif tstate == states.check_fetch_devices:

                        # verify device list of client reflects deamons list
                        tmp = []
                        [ tmp.extend(s.state.dsmids) for s in self.daemons ]

                        if ((len(tmp) != len(cstate.dsmids)) or
                            (set(tmp) != set(cstate.dsmids))):

                            log.info("device list client/deamon mismatch")
                            cstate.dump_device_list();
                            for d in self.daemons:
                                d.dump_device_list()

                            self.assertTrue(False)

                        log.info("device lists match")
                        cstate.dump_device_list();

                        if ct < cutoff:
                            tstate = states.send_new_bc
                            continue

                        self.run = False
                        log.info("")
                        log.info("              SUCCESS")
                        log.info("")

                    continue # otherwise we handle a stale pkt

                #log.debug("evt: %s pkt: %s src: %r", evt,
                #          list2hexstr(pkt) if pkt else None, src)

                if isinstance(src, Connection):

                    if (evt == handle_read):
                        reply = process_pkt(pkt, log, src.state)
                        if pkt[0] == CMD.DS485D_CMD_QUERY_DEVICES:
                            if tstate != states.initialize:
                                delay = random.randint(5, 10)
                                log.info("Daemon%s DS485D_CMD_QUERY_DEVICES delay by %d [s]",
                                        src.state.port, delay)
                                DelayReply(delay, src, reply).start()
                                reply = []

                elif isinstance(src, Client):

                    if (evt == handle_connect):
                        log.info("handle_connect")
                        reply = createQueryBusState(0)
                    elif evt == handle_close:
                        log.info("server closed connection, reconnect")
                        client = None
                    elif (evt == handle_read):
                        # behave like a proper client
                        reply = process_pkt_client(pkt, log, src.state)

                        if pkt[0] == CMD.DS485D_CMD_BUS_CHANGE:
                            log.debug("DS485D_CMD_QUERY_DEVICES started")
                            device_list_queries.append(datetime.datetime.now())
                            reply = createDeviceListQuery()

                        if pkt[0] == CMD.DS485D_CMD_QUERY_DEVICES:
                            log.debug("DS485D_CMD_QUERY_DEVICES received")
                            device_list_replies.append(datetime.datetime.now())

                if not reply:
                    continue

                #log.info("reply: %s",  list2hexstr(reply))
                src.push(bytearray(reply))
                reply = None

            if client:
                client.close()

        except KeyboardInterrupt, e:
            print traceback.format_exc()
            pass
        except AssertionError, e:
            raise
        except Exception, e:
            print traceback.format_exc()
            pass

        log.info("exiting...")

if __name__== "__main__":
    unittest.main()
    #seppuku()
